import { mock, mockClear } from "jest-mock-extended"
import { nanoid } from "nanoid"
import Player from "../lib/Player"
import { getLastEmittedEvent } from "../TestUtils"
import { TownEmitter } from "../types/CoveyTownSocket"
import ConversationArea from "./ConversationArea"


describe('ConversationArea', () => {
    const testAreaBox = { x: 100, y: 100, width: 100, height: 100 };
    let testArea: ConversationArea;
    const townEmitter = mock<TownEmitter>();
    const topic = nanoid();
    const id = nanoid();
    let newPlayer: Player;

    beforeEach(() => {
        mockClear(townEmitter);
        testArea = new ConversationArea({ topic, id, occupantsByID: [] }, testAreaBox, townEmitter);
        newPlayer = new Player(nanoid(), mock<TownEmitter>());
        testArea.add(newPlayer);
    })
    describe('add', () => {
        it('Adds the player to the occupants list and emits an interactableUpdate event', () => {
            expect(testArea.occupantsByID).toEqual([newPlayer.id]);

            const lastEmittedUpdate = getLastEmittedEvent(townEmitter, 'interactableUpdate');
            expect(lastEmittedUpdate).toEqual({ topic, id, occupantsByID: [newPlayer.id] });
        })
        it('Sets the player\'s interactableID and emits an update for their location', () => {
            expect(newPlayer.location.interactableID).toEqual(id);

            const lastEmittedMovement = getLastEmittedEvent(townEmitter, 'playerMoved');
            expect(lastEmittedMovement.location.interactableID).toEqual(id);
        })
    })
    describe('remove', () => {
        it('Removes the player from the list of occupants and emits an interactableUpdate event', () => {
            // Add another player so that we are not also testing what happens when the last player leaves
            const extraPlayer = new Player(nanoid(), mock<TownEmitter>());
            testArea.add(extraPlayer);
            testArea.remove(newPlayer);

            expect(testArea.occupantsByID).toEqual([extraPlayer.id]);
            const lastEmittedUpdate = getLastEmittedEvent(townEmitter, 'interactableUpdate');
            expect(lastEmittedUpdate).toEqual({ topic, id, occupantsByID: [extraPlayer.id] });
        })
        it('Clears the player\'s interactableID and emits an update for their location', () => {
            testArea.remove(newPlayer);
            expect(newPlayer.location.interactableID).toBeUndefined();
            const lastEmittedMovement = getLastEmittedEvent(townEmitter, 'playerMoved');
            expect(lastEmittedMovement.location.interactableID).toBeUndefined();
        })
        it('Clears the topic of the conversation area when the last occupant leaves', () => {
            testArea.remove(newPlayer);
            const lastEmittedUpdate = getLastEmittedEvent(townEmitter, 'interactableUpdate');
            expect(lastEmittedUpdate).toEqual({ topic: undefined, id, occupantsByID: [] });
            expect(testArea.topic).toBeUndefined();
        })
    })
})