import { mock, mockClear } from "jest-mock-extended";
import { nanoid } from "nanoid";
import Player from "../lib/Player";
import { getLastEmittedEvent } from "../TestUtils";
import { TownEmitter } from "../types/CoveyTownSocket";
import ViewingArea from "./ViewingArea";

describe('ViewingArea', () => {
    const testAreaBox = { x: 100, y: 100, width: 100, height: 100 };
    let testArea: ViewingArea;
    const townEmitter = mock<TownEmitter>();
    let newPlayer: Player;
    const id = nanoid();
    const isPlaying = true;
    const elapsedTimeSec = 10;
    const video = nanoid();

    beforeEach(() => {
        mockClear(townEmitter);
        testArea = new ViewingArea({ id, isPlaying, elapsedTimeSec, video }, testAreaBox, townEmitter);
        newPlayer = new Player(nanoid(), mock<TownEmitter>());
        testArea.add(newPlayer);
    })

    describe('remove', () => {
        it('Removes the player from the list of occupants and emits an interactableUpdate event', () => {
            // Add another player so that we are not also testing what happens when the last player leaves
            const extraPlayer = new Player(nanoid(), mock<TownEmitter>());
            testArea.add(extraPlayer);
            testArea.remove(newPlayer);

            expect(testArea.occupantsByID).toEqual([extraPlayer.id]);
            const lastEmittedUpdate = getLastEmittedEvent(townEmitter, 'interactableUpdate');
            expect(lastEmittedUpdate).toEqual({ id, isPlaying, elapsedTimeSec, video });
        })
        it('Clears the player\'s interactableID and emits an update for their location', () => {
            testArea.remove(newPlayer);
            expect(newPlayer.location.interactableID).toBeUndefined();
            const lastEmittedMovement = getLastEmittedEvent(townEmitter, 'playerMoved');
            expect(lastEmittedMovement.location.interactableID).toBeUndefined();
        })
        it('Clears the video property when the last occupant leaves', () => {
            testArea.remove(newPlayer);
            const lastEmittedUpdate = getLastEmittedEvent(townEmitter, 'interactableUpdate');
            expect(lastEmittedUpdate).toEqual({ id, isPlaying, elapsedTimeSec, video: undefined });
            expect(testArea.video).toBeUndefined();
        })
    })
    describe('add', () => {
        it('Adds the player to the occupants list', () => {
            expect(testArea.occupantsByID).toEqual([newPlayer.id]);
        })
        it('Sets the player\'s interactableID and emits an update for their location', () => {
            expect(newPlayer.location.interactableID).toEqual(id);

            const lastEmittedMovement = getLastEmittedEvent(townEmitter, 'playerMoved');
            expect(lastEmittedMovement.location.interactableID).toEqual(id);
        })
    })
})